@@ -1,17 +1,12 @@
 /*
- * Copyright 2002-2003 by Hans Reiser, licensing governed by 
+ * Copyright 2002-2004 by Hans Reiser, licensing governed by 
  * reiserfsprogs/README
  */
 
-#include "includes.h"
+#define _GNU_SOURCE
 
-/* this is provided for anybody who wants to deal with journal */
-int replay_one_transaction (reiserfs_filsys_t *, reiserfs_trans_t *);
-void for_each_transaction (reiserfs_filsys_t *, action_on_trans_t);
-void for_each_block (reiserfs_filsys_t *, reiserfs_trans_t *, action_on_block_t);
-int get_boundary_transactions (reiserfs_filsys_t *, reiserfs_trans_t *,
-			       reiserfs_trans_t *);
-int next_transaction (reiserfs_filsys_t *, reiserfs_trans_t *, reiserfs_trans_t);
+#include "includes.h"
+#include "progbar.h"
 
 /* compares description block with commit block. returns 0 if they differ, 1
    if they match */
@@ -185,7 +180,6 @@ int get_boundary_transactions (reiserfs_filsys_t * fs,
 }
 
 #define TRANS_FOUND     1
-#define TRANS_LAST      2
 #define TRANS_NOT_FOUND 0
 
 /* trans is a valid transaction. Look for valid transaction with smallest
@@ -235,8 +229,6 @@ int next_transaction (reiserfs_filsys_t * fs, reiserfs_trans_t * trans, reiserfs
 	trans->commit_blocknr = commit_expected (fs, next_d_bh);
 	trans->next_trans_offset = next_desc_expected (fs, next_d_bh) - j_start;
 	found = TRANS_FOUND;
-	if (break_trans.trans_id == get_desc_trans_id (next_d_bh))
-	    found = TRANS_LAST;
     }
 
     brelse (d_bh);
@@ -327,6 +319,7 @@ int replay_one_transaction (reiserfs_filsys_t * fs,
 			    reiserfs_trans_t * trans)
 {
     for_each_block (fs, trans, read_journal_write_in_place);
+    fsync(fs->fs_dev);
     return 0;
 }
 
@@ -341,10 +334,7 @@ void for_each_transaction (reiserfs_filsys_t * fs, action_on_trans_t action)
 
     while (1) {
 	action (fs, &oldest);	
-	if (ret == TRANS_LAST)
-	    break;
-	ret = next_transaction (fs, &oldest, newest);	
-	if (ret == TRANS_NOT_FOUND)
+	if ((ret = next_transaction (fs, &oldest, newest)) == TRANS_NOT_FOUND)
 	    break;
     }
 }
@@ -481,10 +471,11 @@ int reiserfs_open_journal (reiserfs_filsys_t * fs, char * j_filename, int flags)
     sb = fs->fs_ondisk_sb;
 
     if (!j_filename) {
+	/*
         if (is_reiserfs_jr_magic_string (sb)) {
-            /* create a special file to access journal */
+            // create a special file to access journal
             return 1;
-        } 
+        } */
 	
 	j_filename = fs->fs_file_name;
     } else if (!is_reiserfs_jr_magic_string (sb)) {
@@ -492,41 +483,56 @@ int reiserfs_open_journal (reiserfs_filsys_t * fs, char * j_filename, int flags)
 	if (strcmp (j_filename, fs->fs_file_name)) {
 	    reiserfs_warning (stderr, "Filesystem with standard journal found, "
 		"wrong name of specified journal device %s \n", j_filename);
-	    return 1;
+	    return 2;
 	}
     }
 
-    fs->fs_journal_dev = open (j_filename, flags | O_LARGEFILE);
+    fs->fs_journal_dev = open (j_filename, flags 
+#if defined(O_LARGEFILE)
+			       | O_LARGEFILE
+#endif
+			       );
     if (fs->fs_journal_dev == -1) 
         return -1;
     
     asprintf (&fs->fs_j_file_name, "%s", j_filename);
     
     if (get_jp_journal_size(sb_jp(sb)) < JOURNAL_MIN_SIZE) {
-	reiserfs_warning (stderr, "Journal of (%lu) block size found on specified journal "
-	    "device %s.\nMust be not less than (%lu).\nRun --rebuild-sb to rebuild journal "
-	    "parameters.\n", get_jp_journal_size (sb_jp (sb)) + 1, j_filename, 
+	reiserfs_warning (stderr, "Journal of (%lu) block size found on "
+	    "specified journal device %s.\nMust be not less than (%lu).\n",
+	    get_jp_journal_size (sb_jp (sb)) + 1, j_filename, 
 	    JOURNAL_MIN_SIZE + 1);
+	close(fs->fs_journal_dev);
 	return 1;
     }
     
-    count = count_blocks (j_filename, fs->fs_blocksize);
-    if (get_jp_journal_1st_block (sb_jp (sb)) + get_jp_journal_size (sb_jp (sb)) + 1 > count) {
-	reiserfs_warning (stderr, "Detected journal on specified device %s does not fit to "
-	    "the device.\nStart block (%lu) + size (%lu) less than device size (%lu).\n"
-	    "Run --rebuild-sb to rebuild journal parameters.\n", 
-	    get_jp_journal_1st_block(sb_jp (sb)), get_jp_journal_size(sb_jp (sb)) + 1,
-	    count);
+    if (!(count = count_blocks (j_filename, fs->fs_blocksize))) {
+	close(fs->fs_journal_dev);
+	return -1;
+    }
+
+    if (get_jp_journal_1st_block (sb_jp (sb)) + 
+	get_jp_journal_size (sb_jp (sb)) + 1 > count) 
+    {
+	reiserfs_warning (stderr, "Detected journal on specified device %s "
+			  "does not fit to the device.\nStart block (%lu) + "
+			  "size (%lu) less than device size (%lu).\n", 
+			  j_filename, get_jp_journal_1st_block(sb_jp (sb)), 
+			  get_jp_journal_size(sb_jp (sb)) + 1, count);
+	close(fs->fs_journal_dev);
 	return 1;
     }
     
     /* read journal header */
-    fs->fs_jh_bh = bread (fs->fs_journal_dev, get_jp_journal_1st_block (sb_jp (sb)) 
-	+ get_jp_journal_size (sb_jp (sb)), fs->fs_blocksize);
+    fs->fs_jh_bh = bread (fs->fs_journal_dev, 
+			  get_jp_journal_1st_block (sb_jp (sb)) + 
+			  get_jp_journal_size (sb_jp (sb)), 
+			  fs->fs_blocksize);
 
     if (!fs->fs_jh_bh) {
-	reiserfs_warning (stderr, "reiserfs_open_journal: bread failed reading "
-	    "journal  header.\n");
+	reiserfs_warning (stderr, "reiserfs_open_journal: bread failed "
+			  "reading journal  header.\n");
+	close(fs->fs_journal_dev);
 	return -1;
     }
 
@@ -540,7 +546,8 @@ int reiserfs_create_journal(
     char * j_device,		/* journal device name */
     unsigned long offset,	/* journal offset on the j_device */
     unsigned long len,		/* including journal header */
-    int transaction_max_size)
+    int transaction_max_size,
+    int force)
 {
     struct stat st;
     struct buffer_head * bh;
@@ -570,7 +577,7 @@ int reiserfs_create_journal(
 	{
 	    /* host device does not contain enough blocks */
 	    reiserfs_warning (stderr, "reiserfs_create_journal: cannot create "
-		"a journal of %lu blocks with %lu offset on %d blocks\n", 
+		"a journal of %lu blocks with %lu offset on %u blocks\n",
 		len, offset, get_sb_block_count(sb));
 		return 0;
 	}
@@ -581,7 +588,9 @@ int reiserfs_create_journal(
 	st.st_rdev = 0;
     } else {
 	/* journal is to be on separate device */
-	blocks = count_blocks (j_device, fs->fs_blocksize);
+	if (!(blocks = count_blocks (j_device, fs->fs_blocksize)))
+		return 0;
+
 	if (!len) {
 	    /* default size of a journal on a separate device is whole device */
 	    if (blocks < offset) {
@@ -589,21 +598,46 @@ int reiserfs_create_journal(
 		    "%lu, blocks on device %lu\n", offset, blocks);
 		return 0;
 	    }
-	    len = blocks - offset;
+	    /* XXX jdm: This can end up being huge and could result
+	     * in an unmountable file system:
+	     * len = blocks - offset; */
+	    len = journal_default_size(fs->fs_super_bh->b_blocknr,
+	                               fs->fs_blocksize) + 1;
+
 	}
 
-	if (len > journal_default_size (fs->fs_super_bh->b_blocknr, 
+	if (!force && len > journal_default_size (fs->fs_super_bh->b_blocknr,
 	    fs->fs_blocksize) + 1) 
 	{
+	    unsigned long journal_max = journal_max_size (fs->fs_super_bh->b_blocknr, fs->fs_blocksize);
 	    fflush(stderr);
 	    
-	    reiserfs_warning (stdout, "NOTE: journal new size %lu is greater "
-		"than default size %lu:\nthis may slow down initializing and "
-		"mounting of the journal. Hope it is ok.\n\n", len, 
-		journal_default_size(fs->fs_super_bh->b_blocknr, 
-		fs->fs_blocksize) + 1);
+	    reiserfs_warning (stdout, "\n*** You've specified a journal "
+	                      "size larger than the default size of "
+	                      "%lu\n*** blocks. This may slow down "
+	                      "journal initialization and mounting "
+	                      "of\n*** the file system.%s",
+	                      journal_default_size(fs->fs_super_bh->b_blocknr, fs->fs_blocksize) + 1,
+	                      len > journal_max ? " " : "\n");
+	    if (len > journal_max)
+		reiserfs_warning (stdout, "On 32-bit systems, and on "
+		                  "64-bit systems with\n*** limited "
+		                  "memory, this may also cause the file "
+		                  "system to be unmountable.\n*** Please "
+		                  "consider using a journal size "
+		                  "<= %lu blocks.\n\nFile system creation "
+				  "failed. You may override this behavior "
+				  "with the -f option.\n", journal_max);
+		    return 0;
 	}
 
+	if (len < blocks)
+		reiserfs_warning (stdout, "\n\n*** Your journal device is %lu "
+		                  "blocks, but your journal is only %lu "
+				  "blocks.\n*** You may want to consider "
+				  "resizing the journal device to avoid "
+				  "wasting space.\n\n", blocks, len);
+
 	if (blocks < offset + len) {
 	    reiserfs_warning (stderr, "reiserfs_create_journal: no enough "
 		"blocks on device %lu, needed %lu\n", blocks, offset + len);
@@ -624,7 +658,11 @@ int reiserfs_create_journal(
 */
     }
 
-    fs->fs_journal_dev = open (j_device, O_RDWR | O_LARGEFILE);
+    fs->fs_journal_dev = open (j_device, O_RDWR 
+#if defined(O_LARGEFILE)
+			       | O_LARGEFILE
+#endif
+			       );
     if (fs->fs_journal_dev == -1) {
 	reiserfs_warning (stderr, "reiserfs_create_journal: could not open "
 	    "%s: %s\n", j_device, strerror(errno));
@@ -633,9 +671,9 @@ int reiserfs_create_journal(
 
     asprintf (&fs->fs_j_file_name, "%s", j_device);
 
-    if (len < JOURNAL_MIN_SIZE) {
+    if (len < JOURNAL_MIN_SIZE + 1) {
 	reiserfs_warning (stderr, "WARNING: Journal size (%u) is less, than "
-	    "minimal supported journal size (%u).", len, JOURNAL_MIN_SIZE + 1);
+	    "minimal supported journal size (%u).\n", len, JOURNAL_MIN_SIZE + 1);
         return 0;
     }
     /* get journal header */
@@ -691,7 +729,11 @@ void reiserfs_reopen_journal (reiserfs_filsys_t * fs, int flag)
     if (close (fs->fs_journal_dev))
 	die ("reiserfs_reopen_journal: closed failed: %s", strerror(errno));
 
-    fs->fs_journal_dev = open (fs->fs_j_file_name, flag | O_LARGEFILE);
+    fs->fs_journal_dev = open (fs->fs_j_file_name, flag 
+#if defined(O_LARGEFILE)
+			       | O_LARGEFILE
+#endif
+			       );
     if (fs->fs_journal_dev == -1)
 	die ("reiserfs_reopen_journal: could not reopen journal device");
 
@@ -734,17 +776,21 @@ void reiserfs_close_journal (reiserfs_filsys_t * fs)
 }
 
 /* update journal header */
-static void update_journal_header (struct buffer_head * bh_jh, reiserfs_trans_t trans) {
+static void update_journal_header (reiserfs_filsys_t * fs, 
+				   struct buffer_head * bh_jh, 
+				   reiserfs_trans_t *trans) 
+{
     struct reiserfs_journal_header * j_head;
 	
     j_head = (struct reiserfs_journal_header *)(bh_jh->b_data);
 
     /* update journal header */
-    set_jh_last_flushed (j_head, trans.trans_id);
-    set_jh_mount_id (j_head, trans.mount_id);
-    set_jh_replay_start_offset (j_head, trans.next_trans_offset);
+    set_jh_last_flushed (j_head, trans->trans_id);
+    set_jh_mount_id (j_head, trans->mount_id);
+    set_jh_replay_start_offset (j_head, trans->next_trans_offset);
     mark_buffer_dirty (bh_jh);
     bwrite (bh_jh);
+    fsync(fs->fs_journal_dev);
 }
 
 /* fixme: what should be done when not all transactions can be replayed in proper order? */
@@ -754,6 +800,8 @@ int replay_journal (reiserfs_filsys_t * fs)
     struct reiserfs_journal_header * j_head;
     reiserfs_trans_t cur, newest, control;
     int replayed, ret;
+    struct progbar progbar;
+    int trans_count;
 
     if (!reiserfs_journal_opened (fs))
         reiserfs_panic ("replay_journal: journal is not opened");
@@ -761,7 +809,8 @@ int replay_journal (reiserfs_filsys_t * fs)
     if (!is_opened_rw (fs))
         reiserfs_panic ("replay_journal: fs is not opened with write perms");
 
-    reiserfs_warning (stderr, "Replaying journal..\n");
+
+    reiserfs_warning (stderr, "Replaying journal: ");
     bh = fs->fs_jh_bh;
 	
     j_head = (struct reiserfs_journal_header *)(bh->b_data);
@@ -774,37 +823,33 @@ int replay_journal (reiserfs_filsys_t * fs)
 	return 0;
     }
 
+    trans_count = newest.trans_id - cur.trans_id;
+
     /*  Smth strange with journal header or journal. We cannot say for sure what was the last 
 	replaied transaction, but relying on JH data is preferable. */
 
     replayed = 0;
     ret = TRANS_FOUND;
     
-    while (cur.mount_id != control.mount_id || cur.trans_id != control.trans_id) {
-	if (control.mount_id == cur.mount_id && control.trans_id == 0 && 
-	    control.desc_blocknr == 0)
+    progbar_init(&progbar, " trans", stderr);
+    /* Looking to the first valid not replayed transaction. */
+    while (1) {
+	if (cur.mount_id == control.mount_id && 
+	    cur.trans_id > control.trans_id)
 	    break;
-	
-	ret = next_transaction (fs, &cur, newest);
-	
-	if (ret != TRANS_FOUND)
+
+	if ((ret = next_transaction (fs, &cur, newest)) != TRANS_FOUND)
 	    break;
     }
-
+    
     while (ret == TRANS_FOUND) {
-	if (control.mount_id != cur.mount_id || control.trans_id != 0 || 
-	    control.desc_blocknr != 0) 
-	{
-	    ret = next_transaction (fs, &cur, newest);
-	    
-	    if (ret == TRANS_NOT_FOUND)
-		break;
-	    
-	    if (cur.mount_id != control.mount_id || cur.trans_id != control.trans_id + 1)
-		break;	    
-	}
+	/* If not the next transaction to be replayed, break out here. */
+	if ((cur.mount_id != control.mount_id) || 
+	    (cur.trans_id != control.trans_id + 1 && control.trans_id))
+	    break;
 	
 	if (!transaction_check_content(fs, &cur)) {
+	    progbar_clear(&progbar);
 	    reiserfs_warning (stderr, "Trans broken: mountid %lu, transid %lu, desc %lu, "
 		"len %lu, commit %lu, next trans offset %lu\n", cur.mount_id, cur.trans_id, 
 		cur.desc_blocknr, cur.trans_len, cur.commit_blocknr, cur.next_trans_offset);
@@ -816,17 +861,28 @@ int replay_journal (reiserfs_filsys_t * fs)
 	    cur.mount_id, cur.trans_id, cur.desc_blocknr, cur.trans_len,
 	    cur.commit_blocknr, cur.next_trans_offset);
 	replay_one_transaction (fs, &cur);
-	update_journal_header (bh, cur);
+	update_journal_header (fs, bh, &cur);
 	control = cur;
         replayed ++;
-    }
 
-    reiserfs_warning (stderr, "%d transactions replayed\n", replayed);
+	progbar_update(&progbar, "Replaying journal", replayed, trans_count,
+	               replayed);
+
+	ret = next_transaction (fs, &cur, newest);
+    }
+    progbar_clear(&progbar);
+
+    reiserfs_warning (stderr, "\rReplaying journal: Done.\nReiserfs journal '%s' in blocks [%u..%u]: %d "
+		      "transactions replayed\n", fs->fs_j_file_name, 
+		      get_jp_journal_1st_block(sb_jp(fs->fs_ondisk_sb)),
+		      get_jp_journal_1st_block(sb_jp(fs->fs_ondisk_sb)) + 
+		      get_jp_journal_size(sb_jp(fs->fs_ondisk_sb)) + 1,
+		      replayed);
 	
     mark_buffer_dirty (fs->fs_super_bh);
     bwrite (fs->fs_super_bh);
 	
-    update_journal_header (bh, newest);
+    update_journal_header (fs, bh, &newest);
 
     return 0;
 }
